/* 
 * Copyright (C) 1996 Szeredi Miklos
 * Email: mszeredi@eik.bme.hu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. See the file COPYING. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "spperif.h"
#include "z80.h"

#include "sptape.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#define MAXFILENAME 1024
#define MAXLINELEN 256

#define SPT_EXT "spt"
#define TAP_EXT "tap"
#define TAP_UEXT "TAP"
#define PAUSE_EXT "PAU"


#define LEADER_IMP 2200
#define SYNC_IMP   700
#define BIT0_IMP   800
#define BIT1_IMP   1600
#define NO_IMP     1000000
#define END_IMP    900


/*
#define LEADER_IMP 2000
#define SYNC_IMP   700
#define BIT0_IMP   700
#define BIT1_IMP   1400
#define NO_IMP     1000000
#define END_IMP    900
*/

static int playing = 0;
static int paused = 0;
static FILE *tapefp;
static FILE *datafile = NULL;
static char tapename[MAXFILENAME];

static int num_segments;
static int playtime;
static int firstbyte;
static int istap;

static int pauselen;
static int line;

static int taplen;

#define PL_NONE    -1
#define PL_NOISE   0
#define PL_LEADER  1
#define PL_DATA    4
#define PL_END     6
#define PL_PAUSE   7

static byte flag;
static char ext[MAXLINELEN];

static char buf[MAXFILENAME];


#define EARBIT 0x40


#define TMSEC(x) (TMNUM * 25 * (x))

#define SPEED_DIV CHKTICK

#define IMPL(x) do {                 \
  imprem = ((x) / SPEED_DIV) * 2;    \
  imphalf = (x) / SPEED_DIV;         \
} while(0)



static int read_tape_line(int real)
{
  int iflag;

  if(fgets(buf, MAXFILENAME, tapefp) == NULL) return 0;

  if(!real) return 1;
  printf("%s", buf);

  sscanf(buf, "%s %x", ext, &iflag);
  flag = (byte) iflag;

  return 1;
}

static void end_segment(void)
{
  if(!istap) {
    if(datafile != NULL) {
      fclose(datafile);
      datafile = NULL;
    }
  }
}

static int seek_next(void)
{
  int c;
  
  c = getc(tapefp);
  if(c == EOF) {
    printf("End of tape\n");
    return 0;
  }
    
  taplen = c;
    
  c = getc(tapefp);
  if(c == EOF) {
    fprintf(stderr, "Error in tape file, incomplete segment\n");
    return 0;
  }
    
  taplen += (c << 8);

  for(; taplen; taplen--) {
    c = getc(tapefp);
    if(c == EOF) {
      fprintf(stderr, "Error in tape file, incomplete segment\n");
      return 0;
    }
  }
  return 1;
}

static int next_segment(void)
{
  static char dataname[MAXFILENAME];

  if(!num_segments) {
    printf("End of playing\n");
    return PL_NONE;
  }

  if(istap) {
    int c;

    c = getc(tapefp);
    if(c == EOF) {
      printf("End of tape\n");
      return PL_NONE;
    }
    
    taplen = c;
    
    c = getc(tapefp);
    if(c == EOF) {
      fprintf(stderr, "Error in tape file, incomplete segment\n");
      return PL_NONE;
    }
    
    taplen += (c << 8);

    c = getc(tapefp);
    taplen--;
    if(c == EOF) {
      fprintf(stderr, "Error in tape file, incomplete segment\n");
      return PL_NONE;
    }
    flag = c;

    printf("Segment size: %i\n", taplen-1);
  }
  else {
    do {
      if(!num_segments) {
	printf("End of playing\n");
	return PL_NONE;
      }
      if(!read_tape_line(1)) {
	if(num_segments < 0) printf("End of tape\n");
	else printf("End of tape directory file found, playing halted\n");
	return PL_NONE;
      }
      
      if(strcmp(ext, PAUSE_EXT) == 0) {
	pauselen = TMSEC(flag);
	return PL_PAUSE;
      }
      
      strcpy(dataname, tapename);
      dataname[strlen(dataname)-4] = '\0';
      strcat(dataname, ".");
      strcat(dataname, ext);
      
      datafile = fopen(dataname, "rb");
      
      if(datafile == NULL) {
	printf("Could not open data file `%s'; %s  Skipping...\n", 
	       dataname, strerror(errno));
	if(num_segments > 0) num_segments--, line++;
      }
    } while(datafile == NULL);
  }

  firstbyte = 1;
  return PL_NOISE;
}

static int next_byte(void)
{
  static int bytectr;
  static byte paritybyte;

  int c;
  
  if(istap) {
    if(firstbyte) {
      bytectr = 0;
      firstbyte = 0;
      return flag;
    }

    if(taplen <= 0) {
      printf("                          \r");
      fflush(stdout);
      return -1;
    }

    c = getc(tapefp);
    if(c == EOF) {
      fprintf(stderr, "Error in tape file, incomplete segment\n");
      return -1;
    }
        
    bytectr++;
    
    if(!(bytectr & 1023)) {
      printf("  at: %i kB                 \r", bytectr >> 10);
      fflush(stdout);
    }

    taplen--;
    return c;
  }
  else {
  

    if(firstbyte) {
      bytectr = 0;
      firstbyte = 0;
      paritybyte = flag;
      return flag;
    }
    
    if(bytectr < 0) {
      printf("                          \r");
      fflush(stdout);
      return -1;
    }
    
    c = getc(datafile);
    if(c == EOF) {
      bytectr = -1;
      return paritybyte;
    }
    
    paritybyte = paritybyte ^ c;
    
    bytectr++;
    
    if(!(bytectr & 1023)) {
      printf("  at: %i kB                 \r", bytectr >> 10);
      fflush(stdout);
    }

    return c;
  }
}

static void stop_playing(void)
{
  if(playing) {
    playing = 0;
    end_segment();
    
    fclose(tapefp);
  }
}


void play_tape()
{
  static int playstate;
  static int imprem;
  static int imphalf;
  static int toput;
  static int bitrem;
  static int starttime;
  int tsl;
  
  if(!playing) {
    if(sp_playing_tape) {
      for(tsl = 0; tsl < TMNUM; tsl++) sp_fe_inport_time[tsl] = 
	SPNM(fe_inport_default);
      sp_playing_tape = 0;
    }
    return;
  }
  if(!sp_playing_tape) sp_playing_tape = 1;

  sp_fe_inport_time[TMNUM] = sp_fe_inport_time[0];

  for(tsl = 0; tsl < TMNUM; tsl++) {
    if(!playtime) {
      playstate = next_segment();
      if(playstate == PL_NONE) {
	stop_playing();
	for(; tsl < TMNUM; tsl++) sp_fe_inport_time[tsl] =
	  SPNM(fe_inport_default);
	return;
      }
      imprem = 0;
    }

    if(!imprem) {
      switch(playstate) {

      case PL_PAUSE:
	if(playtime > pauselen) {
	  playtime = 0;
	  line++;
	  end_segment();
	  continue;
	}
	else IMPL((rand() % NO_IMP) + (NO_IMP / 3));
	break;
	
      case PL_NOISE:
	if(playtime > TMSEC(2)) {
	  playstate = PL_LEADER;
	  starttime = playtime;
	  IMPL(LEADER_IMP);
	}
	else IMPL((rand() % NO_IMP) + (NO_IMP / 3));
	break;
	
      case PL_LEADER:
	if((flag && (playtime - starttime) > TMSEC(2)) 
	   || (playtime - starttime) > TMSEC(3)) {
	  playstate = PL_DATA;
	  bitrem = 0;
	  IMPL(SYNC_IMP);
	}
	else IMPL(LEADER_IMP);
	break;
	
      case PL_DATA:
	if(!bitrem) {
	  toput = next_byte();
	  if(toput < 0) {
	    playstate = PL_END;
	    IMPL(END_IMP);
	    continue;
	  }
	  else bitrem = 8;
	}
	if(toput & 0x80) IMPL(BIT1_IMP);
	else IMPL(BIT0_IMP);
	toput <<= 1;
	bitrem--;
	break;
	
      case PL_END:
	playtime = 0;
	if(num_segments > 0) num_segments--;
	line++;
	end_segment();
	continue;
      }
    }

    if(imprem > imphalf) sp_fe_inport_time[tsl] = 0xFF;
    else sp_fe_inport_time[tsl] = ~EARBIT;
    imprem--;
    playtime++;
  }
}

/*

2168 
2168  (9-10)

667
735   (2-4)

855
855   (3-5)    

1710
1710  (7-9)

945   (4-5)

hosszu: 7..9
rovid:  2..5

*/

#define MICBIT 0x08


#define RC_NONE    0
#define RC_LEADER  1
#define RC_SYNC    2
#define RC_DATA    3

#define MAXLEN      TMNUM

#define LEADER_MIN  9
#define LEADER_MAX  10
#define SYNC_MIN    2
#define SYNC_MAX    4

#define BIT0_MIN    3
#define BIT0_MAX    5
#define BIT1_MIN    7
#define BIT1_MAX    9

#define LEADER_MIN_COUNT 512

static int rec_segment;
static int recording = 0;
static int rec_state = RC_NONE;
static byte *recbuf = NULL;

static const char *waitchars = "-\\|/";


void rec_tape()
{
  static byte lastmic = 0;
  static int lastlen = 0;
  static int whole;
  static int leadercount;
  static byte data;
  static byte parity;
  static int bitnum;
  static int bytecount;
  static int recbufsize;
  static int firsthalf;
  static int frameswait = 0;

  int tsl;
  byte *fep;
  int thishalf;


  if(!recording) return;

  for(fep = sp_fe_outport_time, tsl = TMNUM; tsl; fep++, tsl--) {
    lastlen++;
    if((*fep & MICBIT) == lastmic) {
      if(lastlen < MAXLEN) continue;
    }
    else lastmic = ~lastmic & MICBIT;
    
    switch(rec_state) {
    case RC_NONE:
      if(lastlen >= LEADER_MIN && lastlen <= LEADER_MAX) {
	rec_state = RC_LEADER;

	leadercount = 0;
	break;
      }
      if((frameswait++ & 15)) break;
      frameswait &= 0x3F;
      printf("  %s: WAITING %c            \r", 
	     tapename, waitchars[(frameswait >> 4)]);
      fflush(stdout);
      break;
    case RC_LEADER:
      if(lastlen >= LEADER_MIN && lastlen <= LEADER_MAX) {
	leadercount++;
	if(leadercount == LEADER_MIN_COUNT) {
	  printf("  %s: LEADER            \r", tapename);
	  fflush(stdout);
	}
	break;
      }
      if(leadercount >= LEADER_MIN_COUNT && 
	 lastlen >= SYNC_MIN && lastlen <= SYNC_MAX) rec_state = RC_SYNC;
      else rec_state = RC_NONE;
      break;
    case RC_SYNC:
      if(lastlen >= SYNC_MIN && lastlen <= SYNC_MAX) {
	rec_state = RC_DATA;
	whole = 0;
	data = 0;
	bitnum = 0;
	bytecount = 0;
	recbuf = NULL;
	recbufsize = 0;
	parity = 0;

	printf("  %s: DATA            \r", tapename);
	fflush(stdout);

      }
      else rec_state = RC_NONE;
      break;
    case RC_DATA:
      thishalf = -1;
      if(lastlen >= BIT0_MIN && lastlen <= BIT0_MAX) thishalf = 0;
      else if(lastlen >= BIT1_MIN && lastlen <= BIT1_MAX) thishalf = 1;
      
      if(thishalf < 0 || (whole && firsthalf != thishalf)) {
	char filename[11];
	int filesize;
	int filetype;

	printf("                   \r");

	printf("%s: %03d", tapename, rec_segment);
	if(bytecount >= 1) {
	  printf(" %02X %5d %3s", recbuf[0], bytecount-2, 
		 parity == 0 ? "OK" : "ERR");
	  if(recbuf[0] == 0 && bytecount - 2 >= 17) {
	    filetype = recbuf[1];
	    strncpy(filename, (char*) recbuf+2, 10);
	    filename[10] = '\0';
	    filesize = recbuf[12] + (recbuf[13] << 8);
	    
	    printf("  %02X %10s %5i", filetype, filename, filesize);
	  }
	}
	printf("\n");

	putc(bytecount & 0xFF, tapefp);
	putc((bytecount >> 8) & 0xFF, tapefp);
	
	fwrite(recbuf, 1, bytecount, tapefp);
	fflush(tapefp);

	rec_segment++;
	free(recbuf);
	recbuf = NULL;
	rec_state = RC_NONE;
	break;
      }
      
      if(!whole) {
	whole = 1;
	firsthalf = thishalf;
      }
      else {
	whole = 0;
	data |= thishalf;
	bitnum++;
	
	if(bitnum == 8) {
	  bitnum = 0;
	  if(recbufsize <= bytecount) {
	    recbufsize += 1024;
	    recbuf = realloc(recbuf, recbufsize);
	    if(recbuf == NULL) {
	      fprintf(stderr, "Out of memory\n");
	      exit(1);
	    }
	  }
	  recbuf[bytecount] = data;
	  parity = parity ^ data;
	  data = 0;
	  bytecount++;
	  
	  if(!(bytecount & 1023)) {
	    printf("  %s: DATA %i kB                 \r", 
		   tapename, bytecount >> 10);
	    fflush(stdout);
	  }
	}
	data <<= 1;
      }
      break;
    }

    lastlen = 0;
  }
  
}

static void stop_recording(void)
{
  if(recording) {
    recording = 0;
    free(recbuf);
    recbuf = NULL;

    fclose(tapefp);
  }
}

static void restart_playing(int start_line)
{

  tapefp = fopen(tapename, "rb");
  if(tapefp == NULL) {
    printf("Could not open tape directory file `%s', %s\n", 
	   tapename, strerror(errno));
    return;
  }

  line = 0;
  
  if(istap) {
    while(line < start_line) {
      if(!seek_next()) {
	fclose(tapefp);
	return;
      }
      line++;
    }
  }
  else {
    while(line < start_line) {
      if(!read_tape_line(0)) {
	printf("End of tape directory file `%s' at line %i\n", 
	       tapename, line);
	fclose(tapefp);
	return;
      }
      line++;
    }
  }

  playtime = 0;
  playing = 1;
}

static void start_playing_file(char *name, int start_line, int num)
{
  char *extens;

  strncpy(tapename, name, MAXFILENAME-10);
  tapename[MAXFILENAME-10] = '\0';
  
  extens = tapename+strlen(tapename)-4;
  if(strlen(tapename) < 4 || 
     (strcmp(extens, ".tap") != 0 &&
      strcmp(extens, ".TAP") != 0 &&
      strcmp(extens, ".spt") != 0)) {

    extens = tapename + strlen(tapename);

    if(tapename[0] > 64 && tapename[0] < 96)
      strcat(tapename, "." TAP_UEXT);
    else 
      strcat(tapename, "." TAP_EXT);
  }

  if(strcmp(extens, ".spt") == 0) istap = 0;
  else istap = 1;

  num_segments = num;
  if(start_line < 0) start_line = 0;

  restart_playing(start_line);
}

static void start_recording_file(char *name)
{
  char *extens;
  
  strncpy(tapename, name, MAXFILENAME-10);
  tapename[MAXFILENAME-10] = '\0';
  
  extens = tapename+strlen(tapename)-4;
  if(strlen(tapename) < 4 || 
     (strcmp(extens, ".tap") != 0 &&
      strcmp(extens, ".TAP") != 0)) {
    
    extens = tapename + strlen(tapename);
    
    if(tapename[0] > 64 && tapename[0] < 96)
      strcat(tapename, "." TAP_UEXT);
    else 
      strcat(tapename, "." TAP_EXT);
  }
  
  tapefp = fopen(tapename, "ab");
  if(tapefp == NULL) {
    printf("Could not open tape file `%s', %s\n", 
	   tapename, strerror(errno));
    return;
  }
  
  recording = 1;
  rec_segment = 0;
  rec_state = RC_NONE;

  printf("Recordind tape file `%s'. To stop press Ctrl-s\n", tapename);
}

void pause_play()
{
  if(playing) {
    printf("Paused playing tape `%s' at line %i\n", tapename, line);
     stop_playing();
     paused = 1;
  }
  else if(paused) {
    printf("Continue playing tape `%s' at line %i\n", tapename, line);
    restart_playing(line);
    paused = 0;
  }
}

void start_play()
{
  int start_line, num;
  char *rest, *name;
  int res;

  if(playing || paused || recording) return;

  printf("Enter tape file path (default: '.tap'):\n");

  fgets(buf, MAXFILENAME, stdin);
  buf[MAXFILENAME-1] = '\0';
  
  rest = buf;
  for(; *rest && isspace(*rest); rest++);
  name = rest;
  for(; *rest && isgraph(*rest); rest++);
  
  if(name != rest) res = 1;
  else res = 0;

  if(*rest) {
    *rest = '\0';
    rest++;
    if(*rest) {
      int r1;
      r1 = sscanf(rest, "%d %d", &start_line, &num);
      if(r1 > 0) res += r1;
    }
  }

  if(res < 1) {
    printf("Canceled!\n");
    return;
  }

  if(res < 2) start_line = -1;
  if(res < 3) num = -1;

  start_playing_file(name, start_line, num);
}

void stop_play()
{

  if(playing || paused) {
    printf("Stopped playing tape `%s' at line %i\n", tapename, line);

    if(playing) stop_playing();
    if(paused) paused = 0;
  }
  else if(recording) {
    printf("Stopped recording tape `%s'\n", tapename);
    stop_recording();
  }
}

void start_rec()
{
  char *name, *s;
  
  if(playing || paused || recording) return;

  printf("Enter tape file to record (default: '.tap'):\n");

  fgets(buf, MAXFILENAME, stdin);
  buf[MAXFILENAME-1] = '\0';

  s = buf;
  for(; *s && isspace(*s); s++);
  name = s;
  for(; *s && isgraph(*s); s++);
  *s = '\0';

  if(name == s) {
    printf("Canceled!\n");
    return;
  }
  
  start_recording_file(name);
}

